from math import acos, cos, sin


class ParametricCircle:

    @staticmethod
    def build_circle(radius: float, steps: int) -> list[list[float]]:

        pi2 = acos(-1) * 2.
        points = []
        for step in range(steps):
            a = pi2 / steps * step
            x = radius * cos(a)
            y = radius * sin(a)
            points.append([x, y, 0.])

        return points





from OpenGL.GL import *

from PenguinMol3D.geometries.base_geometry import BaseGeometry
from PenguinMol3D.materials.base_material import BaseMaterial
from PenguinMol3D.objects.base_object_3d import BaseObject3D


class Mesh(BaseObject3D):
    def __init__(self, geometry: BaseGeometry, material: BaseMaterial):
        BaseObject3D.__init__(self)
        self._geometry = geometry
        self._material = material
        self._visible = True

        self._vao_ref = glGenVertexArrays(1)
        glBindVertexArray(self._vao_ref)
        for variable_name, attribute_object in \
            self._geometry.attributes.items():
            attribute_object.associate_variable(material.program_ref,
                                                variable_name)
        glBindVertexArray(0)

    @property
    def geometry(self) -> BaseGeometry:
        return self._geometry

    @property
    def material(self) -> BaseMaterial:
        return self._material

    @property
    def vao_ref(self):
        return self._vao_ref

    @property
    def visible(self) -> bool:
        return self._visible





import numpy as np

from PenguinMol3D.operations.matrix_operations import MatrixOperations


class BaseObject3D:
    def __init__(self):
        self._transform = MatrixOperations.make_identity()
        self._parent = None
        self._children = []

    @property
    def children(self) -> list:
        return self._children

    @property
    def parent(self):
        return self._parent

    @parent.setter
    def parent(self, parent):
        self._parent = parent

    def apply_transform(self,
                        matrix: np.ndarray,
                        local_coord: bool = True
                        ):
        if local_coord:
            self._transform = self._transform @ matrix
        else:
            self._transform = matrix @ self._transform

    def add_child(self, child):
        self._children.append(child)
        child.parent = self

    def get_descendants(self) -> list:
        descendants = []
        nodes_to_process = [self]
        while len(nodes_to_process)>0:
            node = nodes_to_process.pop(0)
            descendants.append(node)
            nodes_to_process = node.children + nodes_to_process
        return descendants

    def get_direction(self) -> list[list]:
        forward = np.array([0.,0.,-1])
        return list(self.get_rotation_transform() @ forward)

    def get_rotation_transform(self) -> np.ndarray:
        return np.array([self._transform[0][0:3],
                         self._transform[1][0:3],
                         self._transform[2][0:3]])

    def get_world_transform(self) -> np.ndarray:
        if self._parent == None:
            return self._transform
        else:
            return self._parent.get_world_transform() @ self._transform

    def get_world_position(self) -> list[float]:
        world_transform = self.get_world_transform()
        return [world_transform.item((0, 3)),
                world_transform.item((1, 3)),
                world_transform.item((2, 3))]

    def get_position(self) -> list[float]:
        return [self._transform.item((0, 3)),
                self._transform.item((1, 3)),
                self._transform.item((2, 3))]

    def look_at(self, target_position: np.ndarray):
        self._transform = MatrixOperations.make_look_at(self.get_world_position(),
                                                        target_position)

    def remove_child(self, child):
        self._children.remove(child)
        child.parent = None

    def rotate_x(self,
                 angle: int | float,
                 local_coord: bool = True):
        m = MatrixOperations.rotate_x(angle)
        self.apply_transform(m, local_coord)

    def rotate_y(self,
                 angle: int | float,
                 local_coord: bool = True):
        m = MatrixOperations.rotate_y(angle)
        self.apply_transform(m, local_coord)

    def rotate_z(self,
                 angle: int | float,
                 local_coord: bool = True):
        m = MatrixOperations.rotate_z(angle)
        self.apply_transform(m, local_coord)

    def set_direction(self, direction):
        position = self.get_position()
        target_position = [position[0] + direction[0],
                           position[1] + direction[1],
                           position[2] + direction[2]]
        self.look_at(target_position)


    def set_position(self, position: list[float]):
        self._transform.itemset((0, 3), position[0])
        self._transform.itemset((1, 3), position[1])
        self._transform.itemset((2, 3), position[2])

    def set_world_transform(self, transform: np.ndarray):
        self._transform = transform

    def scale(self,
              s,
              local_coord: bool = True):
        m = MatrixOperations.scale(s)
        self.apply_transform(m, local_coord)

    def translate(self,
                  x: float,
                  y: float,
                  z: float,
                  local_coord: bool = True):
        m = MatrixOperations.translate(x, y, z)
        self.apply_transform(m, local_coord)






import numpy as np

from PenguinMol3D.general.attribute import Attribute


class BaseShapeGeometry:
    def __init__(self):
        self._attributes = {}
        self._num_vertices = None

    @property
    def attributes(self) -> dict:
        return self._attributes

    @property
    def num_vertices(self) -> int:
        return self._num_vertices

    @num_vertices.setter
    def num_vertices(self, num_vertices: int):
        self._num_vertices = num_vertices

    def add_attribute(self,
                      data: int|float|list,
                      dtype: str,
                      name: str):
        self._attributes[name] = Attribute(data, dtype)

    def count_vertices(self):
        attrib = list(self._attributes.values())[0]
        self._num_vertices = len(attrib.data)

    def apply_transform(self,
                        matrix: np.ndarray,
                        name: str = "vertex_position"):

        old_positions = self._attributes[name].data
        old_vertex_normal_data = self._attributes["vertex_normal"].data
        old_face_normal_data = self._attributes["face_normal"].data

        rotation_matrix = np.array([matrix[0][0:3],
                                    matrix[1][0:3],
                                    matrix[2][0:3]])

        new_positions          = []
        new_vertex_normal_data = []
        new_face_normal_data   = []
        for idx in range(len(old_positions)):
            new_pos         = old_positions[idx].copy()
            new_pos.append(1)
            new_normal      = old_vertex_normal_data[idx].copy()
            new_normal_face = old_face_normal_data[idx].copy()

            new_pos = matrix @ new_pos
            new_pos = list(new_pos[0:3])
            new_normal = rotation_matrix @ new_normal
            new_normal_face = rotation_matrix @ new_normal_face

            new_positions.append(new_pos)
            new_vertex_normal_data.append(new_normal)
            new_face_normal_data.append(new_normal_face)

        self._attributes[name].data = new_positions
        self._attributes["vertex_normal"].data = new_vertex_normal_data
        self._attributes["face_normal"].data = new_face_normal_data

        self._attributes[name].upload_data()

    def merge(self, other_geometry):
        for name, attribute in self._attributes.items():
            attribute.data += other_geometry.attributes[name].data
            attribute.upload_data()

        self.num_vertices += other_geometry.num_vertices


from PenguinMol3D.geometries.atoms.base_atom_geometry import BaseAtomGeometry


class GadoliniumGeometry(BaseAtomGeometry):
    def __new__(cls):
        if not hasattr(cls, "instance"):
            cls.instance = super(GadoliniumGeometry, cls).__new__(cls)
        return cls.instance


    def __init__(self):
        BaseAtomGeometry.__init__(self, radius = 0.54)


from PenguinMol3D.geometries.atoms.base_atom_geometry import BaseAtomGeometry


class HydrogenGeometry(BaseAtomGeometry):
    def __new__(cls):
        if not hasattr(cls, "instance"):
            cls.instance = super(HydrogenGeometry, cls).__new__(cls)
        return cls.instance

    def __init__(self):

        BaseAtomGeometry.__init__(self, radius=0.28)


from PenguinMol3D.general.attribute import Attribute

class BaseGeometryGroup:
    def __init__(self):
        self._group_vertex_position = []
        self._group_vertex_color = []
        self._group_vertex_normal = []
        self._group_face_normal = []
        self._num_vertices = 0
        self._attributes = {}

    @property
    def attributes(self) -> dict:
        return self._attributes

    @property
    def num_vertices(self) -> int:
        return self._num_vertices

    def add_geometry(self, geometry):
        self._group_vertex_position.append(geometry.vertex_position)
        self._group_vertex_color.append(geometry.vertex_color)
        self._group_vertex_normal.append(geometry.vertex_normal)
        self._group_face_normal.append(geometry.face_normal)
        self._num_vertices += geometry.num_vertices

    def create_attributes(self):
        flat_vertex_position = []
        flat_vertex_color = []
        flat_vertex_normal = []
        flat_face_normal = []

        for bond_idx in range(len(self._group_vertex_position)):
            flat_vertex_position.extend(self._group_vertex_position[bond_idx])
            flat_vertex_color.extend(self._group_vertex_color[bond_idx])
            flat_vertex_normal.extend(self._group_vertex_normal[bond_idx])
            flat_face_normal.extend(self._group_face_normal[bond_idx])

        self._attributes["vertex_position"] = Attribute(flat_vertex_position, dtype="vec3")
        self._attributes["vertex_color"] = Attribute(flat_vertex_color, dtype="vec3")
        self._attributes["vertex_normal"] = Attribute(flat_vertex_normal, dtype="vec3")
        self._attributes["face_normal"] = Attribute(flat_face_normal, dtype="vec3")


import numpy as np

class BoundingBox:
    """Container for spatial information"""
    def __init__(self,
                 xmin: float,
                 xmax: float,
                 ymin: float,
                 ymax: float,
                 zmin: float,
                 zmax: float):

        self._x_center_factor = (xmax + xmin) / 2
        self._y_center_factor = (ymax + ymin) / 2
        self._z_center_factor = (zmax + zmin) / 2

        self._width = abs(xmin - xmax)
        self._height = abs(ymin - ymax)
        self._depth = abs(zmin - zmax)

        self._center = [np.median([xmin / self._x_center_factor, xmax / self._x_center_factor]),
                        np.median([ymin / self._y_center_factor, ymax / self._y_center_factor]),
                        np.median([zmin / self._z_center_factor, zmax / self._z_center_factor])]

    @property
    def x_center_factor(self) -> float:
        return self._x_center_factor

    @property
    def y_center_factor(self) -> float:
        return self._y_center_factor

    @property
    def z_center_factor(self) -> float:
        return self._z_center_factor


    @property
    def width(self) -> float:
        return self._width

    @property
    def height(self) -> float:
        return self._height

    @property
    def depth(self) -> float:
        return self._depth


    @property
    def center(self) -> list[float]:
        return self._center





from ctypes import c_void_p

import numpy as np
from OpenGL.GL import *


class Attribute:
    def __init__(self, data: int|float|list, dtype: str):

        self._data = data
        self._dtype = dtype
        self._buffer_ref = glGenBuffers(1)
        self._n_variable = 0

        self.upload_data()

    @property
    def data(self) -> int|float|list:
        return self._data

    @data.setter
    def data(self, data: int | float | list):
        self._data = data

    def associate_variable(self, program_ref, variable_name):
        variable_ref = glGetAttribLocation(program_ref, variable_name)

        if variable_ref == -1:
            return

        glBindBuffer(GL_ARRAY_BUFFER, self._buffer_ref)

        if self._dtype == "int":
            glEnableVertexAttribArray(variable_ref)
            glVertexAttribPointer(variable_ref,1,GL_INT,False,0,None)
        elif self._dtype == "float":
            glEnableVertexAttribArray(variable_ref)
            glVertexAttribPointer(variable_ref,1,GL_FLOAT,False,0,None)
        elif self._dtype == "vec2":
            glEnableVertexAttribArray(variable_ref)
            glVertexAttribPointer(variable_ref,2,GL_FLOAT,False, 0, None)
        elif self._dtype == "vec3":
            glEnableVertexAttribArray(variable_ref)
            glVertexAttribPointer(variable_ref,3,GL_FLOAT,False, 0, None)
        elif self._dtype == "vec4":
            glEnableVertexAttribArray(variable_ref)
            glVertexAttribPointer(variable_ref,4,GL_FLOAT,False, 0, None)
        elif self._dtype == "mat4":
            glEnableVertexAttribArray(variable_ref)
            glEnableVertexAttribArray(variable_ref + 1)
            glEnableVertexAttribArray(variable_ref + 2)
            glEnableVertexAttribArray(variable_ref + 3)

            glVertexAttribPointer(variable_ref, 4, GL_FLOAT, GL_FALSE, 64, c_void_p(0))

            glVertexAttribPointer(variable_ref + 1, 4, GL_FLOAT, GL_FALSE, 64, c_void_p(16))

            glVertexAttribPointer(variable_ref + 2, 4, GL_FLOAT, GL_FALSE, 64, c_void_p(32))

            glVertexAttribPointer(variable_ref + 3, 4, GL_FLOAT, GL_FALSE, 64, c_void_p(48))

            glVertexAttribDivisor(variable_ref, 1)
            glVertexAttribDivisor(variable_ref + 1, 1)
            glVertexAttribDivisor(variable_ref + 2, 1)
            glVertexAttribDivisor(variable_ref + 3, 1)
        else:
            raise Exception(f"Attribute {variable_name} has unknown type {self._dtype}")

    def upload_data(self):
        data = np.array(self._data, dtype=np.float32)
        glBindBuffer(GL_ARRAY_BUFFER, self._buffer_ref)
        if self._dtype == "mat4":
            glBufferData(GL_ARRAY_BUFFER, data.nbytes, data, GL_STATIC_DRAW)
        else:
            glBufferData(GL_ARRAY_BUFFER, data, GL_STATIC_DRAW)



from PenguinMol3D.geometries.shapes.sphere import Sphere

class BaseAtomGeometry(Sphere):

    def __init__(self,
                 radius: int = 1,
                 radius_segments: int = 32,
                 height_segments: int = 16):

        Sphere.__init__(self,
                        radius=radius,
                        radius_segments=radius_segments,
                        height_segments=height_segments)





from PenguinMol3D.geometries.bonds.base_bond_geometry import BaseBond


class TripleBondGeometry(BaseBond):
    def __init__(self,
                 path_points: list[list[float]],
                 path_colors: list[list[float]]):

        BaseBond.__init__(self,
                          path_points,
                          path_colors,
                          contour_radius=0.075,
                          contour_steps=32)


import numpy as np

from PenguinMol3D.operations.matrix_operations import MatrixOperations
from PenguinMol3D.operations.vector_operations import VectorOperations


class Trackball:

    @staticmethod
    def canonicalize_coords(center: np.ndarray, scale: int, point: np.ndarray) -> np.ndarray:
        """Transforms mouse coordinates to canonical space"""
        px = (2 / scale) * (point[0] - center[0])
        py = -(2 / scale) * (point[1] - center[1])
        return np.array([px, py])

    @staticmethod
    def project_onto_sphere(radius: float, point_2d: np.ndarray) -> np.ndarray:
        """Projects an x,y pair onto a sphere of radius r OR a hyperbolic sheet
           if we are away from the center of the sphere.

           Calculates z-axis value based on input 2d point and returns 3d point
           """

        d = np.sqrt(np.dot(point_2d, point_2d))
        if d < radius * 0.70710678118654752440:
            z = np.sqrt(radius * radius - d * d)
            return np.array([point_2d[0], point_2d[1], z])
        else:
            t = radius / 1.41421356237309504880
            return np.array([point_2d[0], point_2d[1], t * t / d])

    @staticmethod
    def simulate_trackball(viewport_size: list[int],
                           curr_pos: list[float],
                           prev_pos: list[float]) -> np.ndarray:

        """ Projects the points onto the virtual trackball, then figures out
            the axis of rotation, which is the cross product of
            P1 P2 and O P1 (O is the center of the ball, 0,0,0)

            Note:  This is a deformed trackball-- is a trackball in the center,
            but is deformed into a hyperbolic sheet of rotation away from the
            center.  This particular function was chosen after trying out
            several variations.

            It is assumed that the arguments to this routine are in the range
            (-1.0 ... 1.0)"""

        # canonicalizes coordinates (-1.0 ... 1.0)
        center = np.array([(viewport_size[0] - 1) / 2, (viewport_size[1] - 1) / 2])
        scale = min(viewport_size[0], viewport_size[1])

        curr_pos = Trackball.canonicalize_coords(center, scale, np.array(curr_pos))
        prev_pos = Trackball.canonicalize_coords(center, scale, np.array(prev_pos))

        # trackball radius
        trackball_radius = 0.8

        # checks zero rotation
        zero_rotation = curr_pos[0] == prev_pos[0] and curr_pos[1] == prev_pos[1]
        if zero_rotation:
            return MatrixOperations.make_identity()

        # calculates 3d coordinates of two 2d points by their projection onto sphere
        curr_pos = Trackball.project_onto_sphere(trackball_radius, curr_pos)
        prev_pos = Trackball.project_onto_sphere(trackball_radius, prev_pos)

        # calculates axis ot rotation
        axis = np.cross(prev_pos, curr_pos)

        # calculates angle of rotation
        t = VectorOperations.calc_vector_length(curr_pos - prev_pos) / (2. * trackball_radius)
        if t >  1.: t =  1.
        if t < -1.: t = -1.
        angle = 2. * np.arcsin(t)

        # calculates quaternion
        quaternion = VectorOperations.calc_quaternion(axis, angle)

        # builds rotational matrix based on quaternion and returns it
        return MatrixOperations.rotate_quaternion(quaternion)





from rdkit.Chem.rdchem import BondType, Mol, Atom

from PenguinMol3D.factories.atoms_factory import AtomsFactory, Atom3D
from PenguinMol3D.factories.bonds_factory import BondsFactory, Bonds3D
from PenguinMol3D.general.bounding_box import BoundingBox
from PenguinMol3D.objects.base_object_3d import BaseObject3D
from PenguinMol3D.operations.matrix_operations import MatrixOperations
from PenguinMol3D.operations.plane_operations import ParametricPlane


class Mol3D(BaseObject3D):
    def __init__(self, mol: Mol):
        BaseObject3D.__init__(self)

        self._mol_rdkit = mol
        self._current_conformer = 0
        self._3d_atoms = {}
        self._3d_bonds = []
        self._bounding_box = None
        self._calc_bounding_box()

        self._generate_3d_model()

    @property
    def atoms(self) -> dict[Atom3D]:
        return self._3d_atoms

    @property
    def bonds(self) -> list[Bonds3D]:
        return self._3d_bonds

    @property
    def bounding_box(self) -> BoundingBox:
        return self._bounding_box

    def _add_bonds(self, bond: Bonds3D):
        self._3d_bonds = bond
        self.add_child(bond)

    def _calc_bounding_box(self):
        """Calculates bounding box of a model"""
        conformer = self._mol_rdkit.GetConformer(self._current_conformer)
        x, y, z = [], [], []
        for position in conformer.GetPositions():
            x.append(position[0])
            y.append(position[1])
            z.append(position[2])

        self._bounding_box = BoundingBox(min(x), max(x), min(y), max(y), min(z), max(z))

    def _translate_to_origin(self, position):

        position[0] -= self._bounding_box.x_center_factor
        position[1] -= self._bounding_box.y_center_factor
        position[2] -= self._bounding_box.z_center_factor


    def _add_atom_transform(self, factory: AtomsFactory, atom_symbol: str, atom_position: list[float]):
        transform = MatrixOperations.make_identity()
        transform.itemset((0, 3), atom_position[0])
        transform.itemset((1, 3), atom_position[1])
        transform.itemset((2, 3), atom_position[2])

        if not atom_symbol in self._3d_atoms:
            self._3d_atoms[atom_symbol] = factory.get_atom_3d(atom_symbol)
            self.add_child(self._3d_atoms[atom_symbol])

        self._3d_atoms[atom_symbol].add_instance(transform)

    def _generate_3d_model(self):
        """Generates 3D model based on molecular data"""

        conformer = self._mol_rdkit.GetConformer(self._current_conformer)
        atoms_factory = AtomsFactory()
        bonds_factory = BondsFactory()

        processed_atoms = []
        for bond in self._mol_rdkit.GetBonds():
            first_atom_idx = bond.GetBeginAtomIdx()
            second_atom_idx = bond.GetEndAtomIdx()

            first_atom = self._mol_rdkit.GetAtomWithIdx(first_atom_idx)
            second_atom = self._mol_rdkit.GetAtomWithIdx(second_atom_idx)

            first_atom_symbol = first_atom.GetSymbol()
            second_atom_symbol = second_atom.GetSymbol()

            first_atom_position = list(conformer.GetAtomPosition(first_atom_idx))
            second_atom_position = list(conformer.GetAtomPosition(second_atom_idx))

            self._translate_to_origin(first_atom_position)
            self._translate_to_origin(second_atom_position)

            if not first_atom_idx in processed_atoms:
                self._add_atom_transform(atoms_factory, first_atom_symbol, first_atom_position)
                processed_atoms.append(first_atom_idx)

            if not second_atom_idx in processed_atoms:
                self._add_atom_transform(atoms_factory, second_atom_symbol, second_atom_position)
                processed_atoms.append(second_atom_idx)

            bond_type = bond.GetBondType()
            normal = None
            if bond_type > BondType.SINGLE:
                third_atom_idx = self._find_third_atom(first_atom, second_atom)
                third_atom_symbol = self._mol_rdkit.GetAtomWithIdx(third_atom_idx).GetSymbol()
                third_atom_position = list(conformer.GetAtomPosition(third_atom_idx))

                self._translate_to_origin(third_atom_position)

                if not third_atom_idx in processed_atoms:
                    self._add_atom_transform(atoms_factory, third_atom_symbol, third_atom_position)
                    processed_atoms.append(third_atom_idx)

                three_atoms_plane = ParametricPlane.from_three_points(first_atom_position,
                                                                      second_atom_position,
                                                                      third_atom_position)
                three_atoms_plane.normalize()
                normal = three_atoms_plane.normal

            bonds_factory.create_bond_geometry([first_atom_position, second_atom_position],
                                                bond_type,
                                                first_atom_symbol,
                                                second_atom_symbol,
                                                normal=normal)


        self._add_bonds(bonds_factory.get_bonds())


    def _find_third_atom(self, atom1: Atom, atom2: Atom) -> int:
        """Finds any other third atom connected to atom1 or atom2 and returns its index"""

        third_atom_bonds = list(atom1.GetBonds())
        this_atom_idx = atom1.GetIdx()
        check_atom_idx = atom2.GetIdx()

        if len(third_atom_bonds) == 1:
            third_atom_bonds = list(atom2.GetBonds())
            this_atom_idx = atom2.GetIdx()
            check_atom_idx = atom1.GetIdx()

        for t_bond in third_atom_bonds:
            other_atom_idx = t_bond.GetOtherAtomIdx(this_atom_idx)
            if other_atom_idx != check_atom_idx:
                return other_atom_idx






from PenguinMol3D.objects.mol_3d import Mol3D
from PenguinMol3D.objects.scene import Scene


class MolecularScene(Scene):
    def __init__(self):
        Scene.__init__(self)
        self._molecules = []

    def add_molecule(self, mol: Mol3D):
        self._molecules.append(mol)
        self.add_child(mol)

    def get_molecules(self) -> list[Mol3D]:
        return self._molecules



import numpy as np

from PenguinMol3D.operations.vector_operations import VectorOperations


class MatrixOperations:

    @staticmethod
    def make_identity() -> np.ndarray:
        return np.array([[1., 0., 0., 0.],
                         [0., 1., 0., 0.],
                         [0., 0., 1., 0.],
                         [0., 0., 0., 1.]],
                        dtype=np.float32)


    @staticmethod
    def make_look_at(position: np.ndarray,
                     target: np.ndarray) -> np.ndarray:
        world_up = [0., 1., 0.]
        forward = np.subtract(target, position)
        right = np.cross(forward, world_up)

        if np.linalg.norm(right) < 0.001:
            offset = np.array([0.001, 0, 0])
            right = np.cross(forward, world_up + offset)

        up = np.cross(right, forward)

        forward = np.divide(forward, np.linalg.norm(forward))
        right = np.divide(right, np.linalg.norm(right))
        up = np.divide(up, np.linalg.norm(up))

        return np.array([[right[0], up[0], -forward[0], position[0]],
                         [right[1], up[1], -forward[1], position[1]],
                         [right[2], up[2], -forward[2], position[2]],
                         [0., 0., 0., 1.]])

    @staticmethod
    def perspective(angle_of_view: int = 60,
                    aspect_ratio: int = 1,
                    near: float = 0.1,
                    far: int = 1000
                    ) -> np.ndarray:
        a = angle_of_view * np.pi / 180.
        d = 1. / np.tan(a / 2)
        r = aspect_ratio
        b = (far + near) / (near - far)
        c = 2 * far * near / (near - far)

        return np.array([[d / r, 0., 0., 0.],
                         [0., d, 0., 0.],
                         [0., 0., b, c],
                         [0., 0., -1., 0.]],
                        dtype=np.float32)

    @staticmethod
    def rotate_align(unit_vec1: np.ndarray, unit_vec2: np.ndarray) -> np.ndarray:
        axis = np.cross(unit_vec1, unit_vec2)
        cos_a = np.dot(unit_vec1, unit_vec2)
        k = 1.0 / 1.0 + cos_a

        return np.array([[axis[0] * axis[0] * k + cos_a,
                          axis[1] * axis[0] * k - axis[2],
                          axis[2] * axis[0] * k + axis[1],
                          0.],
                         [axis[0] * axis[1] * k + axis[2],
                          axis[1] * axis[1] * k + cos_a,
                          axis[2] * axis[1] * k - axis[0],
                          0.],
                         [axis[0] * axis[2] * k - axis[1],
                          axis[1] * axis[2] * k + axis[0],
                          axis[2] * axis[2] * k + cos_a,
                          0.],
                         [0., 0., 0., 1.]])
        """angle = np.arccos(np.dot(unit_vec1, unit_vec2))
        axis = np.cross(unit_vec1, unit_vec2)
        cos = np.cos(angle)
        sin = np.sin(angle)
        t = 1 - cos
        return np.array(  [[cos + axis[0] * axis[0] * t,
                            axis[0] * axis[1] * t - axis[2] * sin,
                            axis[0] * axis[2] * t + axis[1] * sin,
                            0.],
                           [axis[1] * axis[0] * t + axis[2] * sin,
                            cos + (axis[1] * axis[1]) * t,
                            axis[1] * axis[2] * t - axis[0] * sin,
                            0.],
                           [axis[2] * axis[0] * t - axis[1] * sin,
                            axis[2] * axis[1] * t + axis[0] * sin,
                            cos + axis[2] * axis[2] * t,
                            0.],
                           [0., 0., 0., 1.]])"""

    @staticmethod
    def rotate_align_rodrigues(unit_vec1: np.ndarray, unit_vec2: np.ndarray) -> np.ndarray:
        angle = np.arccos(np.dot(unit_vec1, unit_vec2))
        axis = np.cross(unit_vec1, unit_vec2)
        cos = np.cos(angle)
        sin = np.sin(angle)
        t = 1 - cos

        identity = MatrixOperations.make_identity()
        skew = np.array([[ 0.,      -axis[2],  axis[1], 0.],
                         [ axis[2],  0.,      -axis[0], 0.],
                         [-axis[1],  axis[0],  0.     , 0.],
                         [ 0.,       0.,       0.,      1.]])

        return identity + skew + (t * np.dot(skew, skew))

    @staticmethod
    def rotate_align_quaternion(unit_vec1: np.ndarray, unit_vec2: np.ndarray) -> np.ndarray:
        angle = np.arccos(np.dot(unit_vec1, unit_vec2))
        axis = np.cross(unit_vec1, unit_vec2)

        # calculating quaternion
        q = VectorOperations.normalize_quaternion([axis[0] * np.sin(angle / 2),
                                                   axis[1] * np.sin(angle / 2),
                                                   axis[2] * np.sin(angle / 2),
                                                   np.cos(angle/2)])
        qx = q[0]
        qy = q[1]
        qz = q[2]
        qw = q[3]

        #return rotation matrix

        return np.array([[1. - 2. * (qy * qy + qz * qz),
                          2.      * (qx * qy - qz * qw),
                          2.      * (qz * qx + qy * qw),
                          0.],
                         [2.      * (qx * qy + qz * qw),
                          1. - 2. * (qz * qz + qx * qx),
                          2.      * (qy * qz - qx * qw),
                          0.],
                         [2.      * (qz * qx - qy * qw),
                          2.      * (qy * qz + qx * qw),
                          1. - 2. * (qy * qy + qx * qx),
                          0.],
                         [0., 0., 0., 1.]])

    @staticmethod
    def rotate_quaternion(q: np.ndarray | list[float]) -> np.ndarray:
        qx = q[0]
        qy = q[1]
        qz = q[2]
        qw = q[3]

        return np.array([[1. - 2. * (qy * qy + qz * qz),
                          2.      * (qx * qy - qz * qw),
                          2.      * (qz * qx + qy * qw),
                          0.],
                         [2.      * (qx * qy + qz * qw),
                          1. - 2. * (qz * qz + qx * qx),
                          2.      * (qy * qz - qx * qw),
                          0.],
                         [2.      * (qz * qx - qy * qw),
                          2.      * (qy * qz + qx * qw),
                          1. - 2. * (qy * qy + qx * qx),
                          0.],
                         [0., 0., 0., 1.]])


    @staticmethod
    def rotate_x(angle: int) -> np.ndarray:
        c = np.cos(angle)
        s = np.sin(angle)
        return np.array([[1., 0., 0., 0.],
                         [0., c, -s,  0.],
                         [0., s,  c,  0.],
                         [0., 0., 0., 1.]],
                        dtype=np.float32)

    @staticmethod
    def rotate_y(angle: int) -> np.ndarray:
        c = np.cos(angle)
        s = np.sin(angle)
        return np.array([[c,  0., s,  0.],
                         [0., 1., 0., 0.],
                         [-s, 0., c,  0.],
                         [0., 0., 0., 1.]],
                        dtype=np.float32)

    @staticmethod
    def rotate_z(angle: int) -> np.ndarray:
        c = np.cos(angle)
        s = np.sin(angle)
        return np.array([[c, -s,  0., 0.],
                         [s,  c,  0., 0.],
                         [0., 0., 1., 0.],
                         [0., 0., 0., 1.]],
                        dtype=np.float32)

    @staticmethod
    def rotate_axis(axis: np.ndarray, angle: float) -> np.ndarray:
        cos = np.cos(angle)
        sin = np.sin(angle)
        t = 1 - cos
        return np.array([[cos + axis[0] * axis[0] * t,
                          axis[0] * axis[1] * t - axis[2] * sin,
                          axis[0] * axis[2] * t + axis[1] * sin,
                          0.],
                         [axis[1] * axis[0] * t + axis[2] * sin,
                          cos + (axis[1] * axis[1]) * t,
                          axis[1] * axis[2] * t - axis[0] * sin,
                          0.],
                         [axis[2] * axis[0] * t - axis[1] * sin,
                          axis[2] * axis[1] * t + axis[0] * sin,
                          cos + axis[2] * axis[2] * t,
                          0.],
                         [0., 0., 0., 1.]])


    @staticmethod
    def scale(s) -> np.ndarray:
        return np.array([[s,  0., 0., 0.],
                         [0., s,  0., 0.],
                         [0., 0., s,  0.],
                         [0., 0., 0., 1.]],
                        dtype=np.float32)


    @staticmethod
    def scale_nonunifrom(sx: float, sy: float, sz: float) -> np.ndarray:
        return np.array([[sx, 0., 0., 0.],
                         [0., sy, 0., 0.],
                         [0., 0., sz, 0.],
                         [0., 0., 0., 1.]],
                        dtype=np.float32)

    @staticmethod
    def translate(x: float,
                  y: float,
                  z: float
                  ) -> np.ndarray:
        return np.array([[1., 0., 0., x],
                         [0., 1., 0., y],
                         [0., 0., 1., z],
                         [0., 0., 0., 1.]],
                        dtype=np.float32)







import numpy as np

from PenguinMol3D.objects.base_object_3d import BaseObject3D
from PenguinMol3D.operations.matrix_operations import MatrixOperations


class Camera(BaseObject3D):
    def __init__(self,
                 angle_of_view: int = 60,
                 aspect_ratio: int = 1,
                 near: float = 0.1,
                 far: int = 1000):
        BaseObject3D.__init__(self)

        self._angle = angle_of_view
        self._projection = MatrixOperations.perspective\
            (angle_of_view, aspect_ratio, near, far)
        self._view = MatrixOperations.make_identity()

    @property
    def angle(self) -> float:
        return self._angle

    @angle.setter
    def angle(self, angle: float):
        self._angle = angle
    @property
    def projection(self) -> np.ndarray:
        return self._projection

    @property
    def view(self) -> np.ndarray:
        return self._view

    def update_view(self):
        self._view = np.linalg.inv(self.get_world_transform())

    def set_bb_based_position(self, box):
        """Set camera position based on the object spatial information and current angle of view"""
        distance = box.height / np.tan(np.deg2rad(self._angle)/2)
        self.set_position([box.center[0]-1, box.center[1]-1, distance])




import numpy as np

from PenguinMol3D.operations.line_operations import ParametricLine
from PenguinMol3D.operations.matrix_operations import MatrixOperations
from PenguinMol3D.operations.plane_operations import ParametricPlane
from PenguinMol3D.operations.vector_operations import VectorOperations


class ParametricPipe:
    def __init__(self,
                 path_points: list[list[float]],
                 contour_points: list[list[float]]):

        self._path = path_points
        self._contour = contour_points
        self._contours = []
        self._normals = []
        self._generate_contours()

    @property
    def path(self) -> list[list[float]]:
        return self._path

    @path.setter
    def path(self, path: list[list[float]]):
        self._path = path
        self._generate_contours()

    @property
    def contour(self) -> list[list[float]]:
        return self._contour

    @contour.setter
    def contour(self, contour: list[list[float]]):
        self._generate_contours()
        self._contour = contour

    @property
    def contours(self) -> list[list[list[float]]]:
        return self._contours

    @property
    def normals(self) -> list[list[list[float]]]:
        return self._normals

    def _compute_contour_normal(self, path_idx: int) -> list[np.ndarray]:
        """return normal vectors at the current path point"""

        contour = np.array(self._contours[path_idx])
        center = np.array(self._path[path_idx])

        contour_normal = []
        for i in range(len(contour)):
            normal = VectorOperations.calc_unit_vector(contour[i], center)
            contour_normal.append(normal)

        return contour_normal

    def _generate_contours(self):
        """build countour vertex and normal list on each path point"""
        self._contours.clear()
        self._normals.clear()

        if len(self._path) < 1:
            return

        self._transform_first_contour()
        self._contours.append(self._contour)
        self._normals.append(self._compute_contour_normal(0))

        count = len(self._path)
        for i in range(1, count):
            self._contours.append(self._project_contour(i - 1, i))
            self._normals.append(self._compute_contour_normal(i))
        #print(self._contours)

    def _project_contour(self, from_idx: int, to_idx: int):
        """project a contour to a plane at the path point"""
        dir1 = np.array(self._path[to_idx]) \
               - np.array(self._path[from_idx])
        dir2 = None
        if to_idx == len(self._path) - 1:
            dir2 = dir1
        else:
            dir2 = np.array(self._path[to_idx + 1]) \
                   - np.array(self._path[to_idx])

        normal = dir1 + dir2
        plane = ParametricPlane(self._path[to_idx], normal)

        from_contour = self._contours[from_idx]

        to_contour = []
        for i in range(len(from_contour)):
            line = ParametricLine(dir1, from_contour[i])
            to_contour.append(plane.get_intersection_line(line))

        return to_contour

    def _transform_first_contour(self):
        path_count = len(self._path)
        contour_count = len(self._contour)

        transform = MatrixOperations.make_identity()
        p1 = np.array(self._path[0])
        p2 = np.array(self._path[1])
        if path_count:
            transform = transform @ MatrixOperations.translate(p1[0], p1[1], p1[2])

            if path_count > 1:
                transform = MatrixOperations.make_look_at(p1, p2)

            for i in range(contour_count):
                vector = self._contour[i]
                vector.append(1)
                self._contour[i] = np.dot(transform, vector)[:3]


    def add_path_point(self, point: list[float]):
        self._path.append(point)

        count = len(self._path)
        if count == 1:
            self._transform_first_contour()
            self._normals.append(self._compute_contour_normal(0))

        elif count == 2:
            self._contours.append(self._project_contour(0, 1))
            self._normals.append(self._compute_contour_normal(1))
        else:
            dummy = []
            self._contours.append(dummy)
            self._normals.append(dummy)

            self._contours[count - 2] = self._project_contour(count - 3, count - 2)
            self._normals[count - 2] = self._compute_contour_normal(count - 2)

            self._contours[count - 1] = self._project_contour(count - 2, count - 1)
            self._normals[count - 1] = self._compute_contour_normal(count - 1)













from OpenGL.GL import *

from PenguinMol3D.general.opengl_utils import OpenGLUtils
from PenguinMol3D.general.uniform import Uniform


class BaseMaterial:
    """Base material class"""
    def __init__(self, vert_shader_code: str, frag_shader_code: str):

        self._program_ref = OpenGLUtils.initialize_program(vert_shader_code,
                                                           frag_shader_code)
        self._uniforms = {}
        #self._uniforms["model_matrix"]     = Uniform(None, "mat4")
        self._uniforms["view_matrix"]       = Uniform(None, "mat4")
        self._uniforms["projection_matrix"] = Uniform(None, "mat4")

        self._settings = {}
        self._settings["draw_style"] = GL_TRIANGLES

    @property
    def program_ref(self):
        return self._program_ref

    @property
    def uniforms(self) -> dict:
        return self._uniforms

    @property
    def settings(self) -> dict:
        return self._settings

    def add_uniform(self, data: int|float|list, dtype: str, name: str):
        self._uniforms[name] = Uniform(data, dtype)

    def locate_uniform(self, name: str):
        self._uniforms[name].locate_variable(self._program_ref, name)

    def locate_uniforms(self):
        for name, object in self._uniforms.items():
            object.locate_variable(self._program_ref, name)

    def update_render_settings(self):
        pass

    def set_properties(self, properties: dict):
        for name, data in properties.items():
            if name in self._uniforms:
                self._uniforms[name].data = data
            elif name in self._settings:
                self._settings[name] = data
            else:
                raise Exception("Material has no property named:" + name)







import numpy as np


class BaseGeometry:
    def __init__(self):
        self._vertex_color = []
        self._vertex_position = []
        self._vertex_normal = []
        self._face_normal = []
        self._num_vertices = 0

    @property
    def vertex_position(self) -> list:
        return self._vertex_position

    @vertex_position.setter
    def vertex_position(self, new_vertex_position: list):
        self._vertex_position = new_vertex_position

    @property
    def vertex_color(self) -> list:
        return self._vertex_color

    @vertex_color.setter
    def vertex_color(self, new_vertex_color: list):
        self._vertex_color = new_vertex_color

    @property
    def vertex_normal(self) -> list:
        return self._vertex_normal

    @vertex_normal.setter
    def vertex_normal(self, new_vertex_normal: list):
        self._vertex_normal = new_vertex_normal

    @property
    def face_normal(self) -> list:
        return self._face_normal

    @face_normal.setter
    def face_normal(self, new_face_normal: list):
        self._face_normal = new_face_normal

    @property
    def num_vertices(self) -> int:
        return self._num_vertices

    @num_vertices.setter
    def num_vertices(self, new_num_vertices: int):
        self._num_vertices = new_num_vertices

    def apply_transform(self,
                        matrix: np.ndarray):

        rotation_matrix = np.array([matrix[0][0:3],
                                    matrix[1][0:3],
                                    matrix[2][0:3]])
        new_positions = []
        new_vertex_normal_data = []
        new_face_normal_data = []
        for idx in range(len(self.vertex_position)):
            new_pos = self.vertex_position[idx].copy()
            new_pos.append(1)
            new_normal = self.vertex_normal[idx].copy()
            #new_normal_face = self.face_normal[idx].copy()

            new_pos = matrix @ new_pos
            new_pos = list(new_pos[0:3])
            new_normal = rotation_matrix @ new_normal
            #new_normal_face = rotation_matrix @ new_normal_face

            new_positions.append(new_pos)
            new_vertex_normal_data.append(new_normal)
            #new_face_normal_data.append(new_normal_face)

        self.vertex_position = new_positions
        self.vertex_normal = new_vertex_normal_data
        self.face_normal = new_face_normal_data

    def merge(self, other_bond):
        self.vertex_position += other_bond.vertex_position
        self.vertex_color += other_bond.vertex_color
        self.vertex_normal += other_bond.vertex_normal
        self.face_normal += other_bond.face_normal
        self.num_vertices += other_bond.num_vertices


import numpy as np


class ParametricLine:

    def __init__(self, direction: np.ndarray, point: np.ndarray):
        """class to construct a line with parametric form"""

        self.direction = direction
        self.point = np.array(point)

    def __repr__(self) -> str:
        return f"ParametricLine (start: {self.point}, " \
               f"direction: {self.direction})"

    @classmethod
    def from_points(cls, p1: list[float], p2: list[float]):
        point = np.array(p1)
        direction = np.array(p2) - p1
        return cls(direction, point)


    def get_normalized_direction(self) -> np.ndarray:
        """returns unit vector representing line direction"""
        return self.direction / np.linalg.norm(self.direction)


    def get_intersection(self, line) -> np.ndarray:
        """returns the intersection point with the other line
           or NaN if lines are parallel"""

        d2 = line.direction
        p2 = line.point

        d3 = np.cross(p2 - self.point, d2)
        d4 = np.cross(self.direction, d2)

        dot = np.dot(d4, d4)
        if not dot.any():
            return np.array([np.nan, np.nan, np.nan])

        alpha = np.dot(d3, d4) / dot

        return self.point + (alpha * self.direction)


    def is_intersected(self, line) -> bool:
        """determine if this line intersects with the other line"""
        vector = np.cross(self.direction,
                          line.direction)
        if not vector.any():
            return False
        return True


from PenguinMol3D.objects.light.base_light import BaseLight


class Point(BaseLight):
    def __init__(self,
                 color: list[float] = [1., 1., 1.],
                 position: list[float] = [1., 1., 1.],
                 attenuation: list[float] = [1., 0., 0., 1.]):
        BaseLight.__init__(self, light_type=BaseLight.POINT)
        self.color = color
        self.set_position(position)
        self.attenuation = attenuation


from OpenGL.GL import *
from PIL.Image import frombytes

from PenguinMol3D.objects.camera import Camera
from PenguinMol3D.objects.light.base_light import BaseLight
from PenguinMol3D.objects.molecular_scene import MolecularScene


class Renderer:
    def __init__(self, clear_color: list[float] = [0., 0. ,0.]):
        glEnable(GL_DEPTH_TEST)
        glEnable(GL_MULTISAMPLE)
        glClearColor(clear_color[0], clear_color[1], clear_color[2], 1)

    def _render_lights(self, lights, mesh):
        if "light0" in mesh.material.uniforms.keys():
            for light_number in range(4):
                light_name = f"light{str(light_number)}"
                light_object = lights[light_number]
                mesh.material.uniforms[light_name].data = light_object

    def _render_atoms(self, atoms, lights, camera_pos, camera_view, camera_proj):
        for atom in atoms:
            glUseProgram(atom.material.program_ref)
            glBindVertexArray(atom.vao_ref)

            instances = atom.get_instances()
            for no, instance in enumerate(instances):
                atom.material.uniforms[f"model_matrix[{no}]"].data = atom.get_world_transform() @ instance

            atom.material.uniforms["view_matrix"].data = camera_view
            atom.material.uniforms["projection_matrix"].data = camera_proj

            if "view_position" in atom.material.uniforms.keys():
                atom.material.uniforms["view_position"].data = camera_pos

            self._render_lights(lights, atom)

            for uniform in atom.material.uniforms.values():
                uniform.upload_data()

            atom.material.update_render_settings()

            glDrawArraysInstanced(atom.material.settings["draw_style"],
                                  0,
                                  atom.geometry.num_vertices,
                                  len(instances))

            glBindVertexArray(0)

    def _render_bonds(self, bonds, lights, camera_pos, camera_view, camera_proj):
        glUseProgram(bonds.material.program_ref)
        glBindVertexArray(bonds.vao_ref)

        bonds.material.uniforms["model_matrix"].data = bonds.get_world_transform()
        bonds.material.uniforms["view_matrix"].data = camera_view
        bonds.material.uniforms["projection_matrix"].data = camera_proj

        if "view_position" in bonds.material.uniforms.keys():
            bonds.material.uniforms["view_position"].data = camera_pos

        self._render_lights(lights, bonds)

        for uniform in bonds.material.uniforms.values():
            uniform.upload_data()

        bonds.material.update_render_settings()

        glDrawArrays(bonds.material.settings["draw_style"],0,bonds.geometry.num_vertices)
        glBindVertexArray(0)

    def render(self, scene: MolecularScene, camera: Camera):
        glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT)
        camera.update_view()

        descendants = scene.get_descendants()

        light_filter = lambda x: isinstance(x, BaseLight)
        lights = list(filter(light_filter, descendants))

        while len(lights) < 4:
            lights.append(BaseLight())

        camera_pos  = camera.get_world_position()
        camera_view = camera.view
        camera_proj = camera.projection

        molecule = scene.get_molecules()[0]

        self._render_atoms(molecule.atoms.values(), lights, camera_pos, camera_view, camera_proj)
        self._render_bonds(molecule.bonds, lights, camera_pos, camera_view, camera_proj)

    def save_frame(self, width: int, height: int, filepath: str):
        content = glReadPixels(0, 0, width, height, GL_RGB, GL_UNSIGNED_BYTE)
        image = frombytes("RGB", (width, height), content, "raw", "RGB", 0, 1)
        image = image.transpose(1)
        image.save(filepath)









from PenguinMol3D.general.globals import ELEMENT_COLORS
from PenguinMol3D.geometries.atoms import (
    hydrogen_geometry,
    helium_geometry,
    lithium_geometry,
    beryllium_geometry,
    boron_geometry,
    carbon_geometry,
    nitrogen_geometry,
    oxygen_geometry,
    fluorine_geometry,
    neon_geometry,
    sodium_geometry,
    magnesium_geometry,
    aluminum_geometry,
    silicon_geometry,
    phosphorus_geometry,
    sulfur_geometry,
    chlorine_geometry,
    argon_geometry,
    potassium_geometry,
    calcium_geometry,
    scandium_geometry,
    titanium_geometry,
    vanadium_geometry,
    manganese_geometry,
    iron_geometry,
    cobalt_geometry,
    nickel_geometry,
    copper_geometry,
    zinc_geometry,
    gallium_geometry,
    germanium_geometry,
    arsenic_geometry,
    selenium_geometry,
    bromine_geometry,
    krypton_geometry,
    rubidium_geometry,
    strontium_geometry,
    yttrium_geometry,
    zirconium_geometry,
    niobium_geometry,
    molybdenum_geometry,
    technetium_geometry,
    ruthenium_geometry,
    rhodium_geometry,
    palladium_geometry,
    silver_geometry,
    cadmium_geometry,
    indium_geometry,
    tin_geometry,
    antimony_geometry,
    tellurium_geometry,
    iodine_geometry,
    xenon_geometry,
    caesium_geometry,
    barium_geometry,
    lanthanum_geometry,
    cerium_geometry,
    praseodymium_geometry,
    neodymium_geometry,
    promethium_geometry,
    samarium_geometry,
    europium_geometry,
    gadolinium_geometry,
    terbium_geometry,
    dysprosium_geometry,
    holmium_geometry,
    erbium_geometry,
    thulium_geometry,
    ytterbium_geometry,
    lutetium_geometry,
    hafnium_geometry,
    tantalum_geometry,
    tungsten_geometry,
    rhenium_geometry,
    osmium_geometry,
    irirdium_geometry,
    platinum_geometry,
    gold_geometry,
    mercury_geometry,
    thallium_geometry,
    bismuth_geometry,
    polonium_geometry,
    astatine_geometry,
    radon_geometry,
    francium_geometry,
    radium_geometry,
    actinium_geometry,
    thorium_geometry,
    protactinium_geometry,
    uranium_geometry,
    neptunium_geometry,
    plutonium_geometry,
    americium_geometry,
    curium_geometry,
    berkelium_geometry,
    californium_geometry,
    einsteinium_geometry
)

from PenguinMol3D.materials.atom_phong_material import PhongMaterial
from PenguinMol3D.objects.mesh import Mesh

class Atom3D(Mesh):
    def __init__(self, geometry, material):
        self._instances = []
        Mesh.__init__(self, geometry, material)

    def get_instances(self) -> list:
        return self._instances

    def add_instance(self, instance):
        self._instances.append(instance)
        name = f"model_matrix[{len(self._instances)-1}]"
        self.material.add_uniform(instance, "mat4", name)
        self.material.locate_uniform(name)

class AtomsFactory:
    def __init__(self):
        self._atoms_parameters = {}

    def get_atom_3d(self, atom_symbol: str) -> Atom3D:
        atom_3d = None

        if atom_symbol not in self._atoms_parameters:
            material = PhongMaterial(properties={"base_color": ELEMENT_COLORS[atom_symbol]})
            material.settings["cull_face"] = True
            material.settings["back_side"] = False
            if   atom_symbol == "H":
                self._atoms_parameters[atom_symbol] = (hydrogen_geometry.HydrogenGeometry(), material)
            elif atom_symbol == "He":
                self._atoms_parameters[atom_symbol] = (helium_geometry.HeliumGeometry(), material)
            elif atom_symbol == "Li":
                self._atoms_parameters[atom_symbol] = (lithium_geometry.LithiumGeometry(), material)
            elif atom_symbol == "Be":
                self._atoms_parameters[atom_symbol] = (beryllium_geometry.BerylliumGeometry(), material)
            elif atom_symbol == "B":
                self._atoms_parameters[atom_symbol] = (boron_geometry.BoronGeometry(), material)
            elif atom_symbol == "C":
                self._atoms_parameters[atom_symbol] = (carbon_geometry.CarbonGeometry(), material)
            elif atom_symbol == "N":
                self._atoms_parameters[atom_symbol] = (nitrogen_geometry.NitrogenGeometry(), material)
            elif atom_symbol == "O":
                self._atoms_parameters[atom_symbol] = (oxygen_geometry.OxygenGeometry(), material)
            elif atom_symbol == "F":
                self._atoms_parameters[atom_symbol] = (fluorine_geometry.FluorineGeometry(), material)
            elif atom_symbol == "Ne":
                self._atoms_parameters[atom_symbol] = (neon_geometry.NeonGeometry(), material)
            elif atom_symbol == "Na":
                self._atoms_parameters[atom_symbol] = (sodium_geometry.SodiumGeometry(), material)
            elif atom_symbol == "Mg":
                self._atoms_parameters[atom_symbol] = (magnesium_geometry.MagnesiumGeometry(), material)
            elif atom_symbol == "Al":
                self._atoms_parameters[atom_symbol] = (aluminum_geometry.AluminumGeometry(), material)
            elif atom_symbol == "Si":
                self._atoms_parameters[atom_symbol] = (silicon_geometry.SiliconGeometry(), material)
            elif atom_symbol == "P":
                self._atoms_parameters[atom_symbol] = (phosphorus_geometry.PhosphorusGeometry(), material)
            elif atom_symbol == "S":
                self._atoms_parameters[atom_symbol] = (sulfur_geometry.SulfurGeometry(), material)
            elif atom_symbol == "Cl":
                self._atoms_parameters[atom_symbol] = (chlorine_geometry.ChlorineGeometry(), material)
            elif atom_symbol == "Ar":
                self._atoms_parameters[atom_symbol] = (argon_geometry.ArgonGeometry(), material)
            elif atom_symbol == "K":
                self._atoms_parameters[atom_symbol] = (potassium_geometry.PotassiumGeometry(), material)
            elif atom_symbol == "Ca":
                self._atoms_parameters[atom_symbol] = (calcium_geometry.CalciumGeometry(), material)
            elif atom_symbol == "Sc":
                self._atoms_parameters[atom_symbol] = (scandium_geometry.ScandiumGeometry(), material)
            elif atom_symbol == "Ti":
                self._atoms_parameters[atom_symbol] = (titanium_geometry.TitaniumGeometry(), material)
            elif atom_symbol == "V":
                self._atoms_parameters[atom_symbol] = (vanadium_geometry.VanadiumGeometry(), material)
            elif atom_symbol == "Cr":
                self._atoms_parameters[atom_symbol] = (vanadium_geometry.ChromiumGeometry(), material)
            elif atom_symbol == "Mn":
                self._atoms_parameters[atom_symbol] = (manganese_geometry.ManganeseGeometry(), material)
            elif atom_symbol == "Fe":
                self._atoms_parameters[atom_symbol] = (iron_geometry.IronGeometry(), material)
            elif atom_symbol == "Co":
                self._atoms_parameters[atom_symbol] = (cobalt_geometry.CobaltGeometry(), material)
            elif atom_symbol == "Ni":
                self._atoms_parameters[atom_symbol] = (nickel_geometry.NickelGeometry(), material)
            elif atom_symbol == "Cu":
                self._atoms_parameters[atom_symbol] = (copper_geometry.CopperGeometry(), material)
            elif atom_symbol == "Zn":
                self._atoms_parameters[atom_symbol] = (zinc_geometry.ZincGeometry(), material)
            elif atom_symbol == "Ga":
                self._atoms_parameters[atom_symbol] = (gallium_geometry.GalliumGeometry(), material)
            elif atom_symbol == "Ge":
                self._atoms_parameters[atom_symbol] = (germanium_geometry.GermaniumGeometry(), material)
            elif atom_symbol == "As":
                self._atoms_parameters[atom_symbol] = (arsenic_geometry.ArsenicGeometry(), material)
            elif atom_symbol == "Se":
                self._atoms_parameters[atom_symbol] = (selenium_geometry.SeleniumGeometry(), material)
            elif atom_symbol == "Br":
                self._atoms_parameters[atom_symbol] = (bromine_geometry.BromineGeometry(), material)
            elif atom_symbol == "Kr":
                self._atoms_parameters[atom_symbol] = (krypton_geometry.KryptonGeometry(), material)
            elif atom_symbol == "Rb":
                self._atoms_parameters[atom_symbol] = (rubidium_geometry.RubidiumGeometry(), material)
            elif atom_symbol == "Sr":
                self._atoms_parameters[atom_symbol] = (strontium_geometry.StrontiumGeometry(), material)
            elif atom_symbol == "Y":
                self._atoms_parameters[atom_symbol] = (yttrium_geometry.YttriumGeometry(), material)
            elif atom_symbol == "Zr":
                self._atoms_parameters[atom_symbol] = (zirconium_geometry.ZirconiumGeometry(), material)
            elif atom_symbol == "Nb":
                self._atoms_parameters[atom_symbol] = (niobium_geometry.NiobiumGeometry(), material)
            elif atom_symbol == "Mo":
                self._atoms_parameters[atom_symbol] = (molybdenum_geometry.MolybdenumGeometry(), material)
            elif atom_symbol == "Tc":
                self._atoms_parameters[atom_symbol] = (technetium_geometry.TechnetiumGeometry(), material)
            elif atom_symbol == "Ru":
                self._atoms_parameters[atom_symbol] = (ruthenium_geometry.RutheniumGeometry(), material)
            elif atom_symbol == "Rh":
                self._atoms_parameters[atom_symbol] = (rhodium_geometry.RhodiumGeometry(), material)
            elif atom_symbol == "Pd":
                self._atoms_parameters[atom_symbol] = (palladium_geometry.PalladiumGeometry(), material)
            elif atom_symbol == "Ag":
                self._atoms_parameters[atom_symbol] = (silver_geometry.SilverGeometry(), material)
            elif atom_symbol == "Cd":
                self._atoms_parameters[atom_symbol] = (cadmium_geometry.CadmiumGeometry(), material)
            elif atom_symbol == "In":
                self._atoms_parameters[atom_symbol] = (indium_geometry.IndiumGeometry(), material)
            elif atom_symbol == "Sn":
                self._atoms_parameters[atom_symbol] = (tin_geometry.TinGeometry(), material)
            elif atom_symbol == "Sb":
                self._atoms_parameters[atom_symbol] = (antimony_geometry.AntimonyGeometry(), material)
            elif atom_symbol == "Te":
                self._atoms_parameters[atom_symbol] = (tellurium_geometry.TelluriumGeometry(), material)
            elif atom_symbol == "I":
                self._atoms_parameters[atom_symbol] = (iodine_geometry.IodineGeometry(), material)
            elif atom_symbol == "Xe":
                self._atoms_parameters[atom_symbol] = (xenon_geometry.XenonGeometry(), material)
            elif atom_symbol == "Cs":
                self._atoms_parameters[atom_symbol] = (caesium_geometry.CaesiumGeometry(), material)
            elif atom_symbol == "Ba":
                self._atoms_parameters[atom_symbol] = (barium_geometry.BariumGeometry(), material)
            elif atom_symbol == "La":
                self._atoms_parameters[atom_symbol] = (lanthanum_geometry.LanthanumGeometry(), material)
            elif atom_symbol == "Ce":
                self._atoms_parameters[atom_symbol] = (cerium_geometry.CeriumGeometry(), material)
            elif atom_symbol == "Pr":
                self._atoms_parameters[atom_symbol] = (praseodymium_geometry.PraseodymiumGeometry(), material)
            elif atom_symbol == "Nd":
                self._atoms_parameters[atom_symbol] = (neodymium_geometry.NeodymiumGeometry(), material)
            elif atom_symbol == "Pm":
                self._atoms_parameters[atom_symbol] = (promethium_geometry.PromethiumGeometry(), material)
            elif atom_symbol == "Sm":
                self._atoms_parameters[atom_symbol] = (samarium_geometry.SamariumGeometry(), material)
            elif atom_symbol == "Eu":
                self._atoms_parameters[atom_symbol] = (europium_geometry.EuropiumGeometry(), material)
            elif atom_symbol == "Gd":
                self._atoms_parameters[atom_symbol] = (gadolinium_geometry.GadoliniumGeometry(), material)
            elif atom_symbol == "Tb":
                self._atoms_parameters[atom_symbol] = (terbium_geometry.TerbiumGeometry(), material)
            elif atom_symbol == "Dy":
                self._atoms_parameters[atom_symbol] = (dysprosium_geometry.DysprosiumGeometry(), material)
            elif atom_symbol == "Ho":
                self._atoms_parameters[atom_symbol] = (holmium_geometry.HolmiumGeometry(), material)
            elif atom_symbol == "Er":
                self._atoms_parameters[atom_symbol] = (erbium_geometry.ErbiumGeometry(), material)
            elif atom_symbol == "Th":
                self._atoms_parameters[atom_symbol] = (thulium_geometry.ErbiumGeometry(), material)
            elif atom_symbol == "Yb":
                self._atoms_parameters[atom_symbol] = (ytterbium_geometry.YtterbiumGeometry(), material)
            elif atom_symbol == "Lu":
                self._atoms_parameters[atom_symbol] = (lutetium_geometry.LutetiumGeometry(), material)
            elif atom_symbol == "Hf":
                self._atoms_parameters[atom_symbol] = (hafnium_geometry.HafniumGeometry(), material)
            elif atom_symbol == "Ta":
                self._atoms_parameters[atom_symbol] = (tantalum_geometry.TantalumGeometry(), material)
            elif atom_symbol == "W":
                self._atoms_parameters[atom_symbol] = (tungsten_geometry.TungstenGeometry(), material)
            elif atom_symbol == "Re":
                self._atoms_parameters[atom_symbol] = (rhenium_geometry.RheniumGeometry(), material)
            elif atom_symbol == "Os":
                self._atoms_parameters[atom_symbol] = (osmium_geometry.OsmiumGeometry(), material)
            elif atom_symbol == "Ir":
                self._atoms_parameters[atom_symbol] = (irirdium_geometry.IridiumGeometry(), material)
            elif atom_symbol == "Pt":
                self._atoms_parameters[atom_symbol] = (platinum_geometry.PlatinumGeometry(), material)
            elif atom_symbol == "Au":
                self._atoms_parameters[atom_symbol] = (gold_geometry.GoldGeometry(), material)
            elif atom_symbol == "Hg":
                self._atoms_parameters[atom_symbol] = (mercury_geometry.MercuryGeometry(), material)
            elif atom_symbol == "Tl":
                self._atoms_parameters[atom_symbol] = (thallium_geometry.ThalliumGeometry(), material)
            elif atom_symbol == "Bi":
                self._atoms_parameters[atom_symbol] = (bismuth_geometry.BismuthGeometry(), material)
            elif atom_symbol == "Po":
                self._atoms_parameters[atom_symbol] = (polonium_geometry.PoloniumGeometry(), material)
            elif atom_symbol == "At":
                self._atoms_parameters[atom_symbol] = (astatine_geometry.AstatineGeometry(), material)
            elif atom_symbol == "Rn":
                self._atoms_parameters[atom_symbol] = (radon_geometry.RadonGeometry(), material)
            elif atom_symbol == "Fr":
                self._atoms_parameters[atom_symbol] = (francium_geometry.FranciumGeometry(), material)
            elif atom_symbol == "Ra":
                self._atoms_parameters[atom_symbol] = (radium_geometry.RadiumGeometry(), material)
            elif atom_symbol == "Ac":
                self._atoms_parameters[atom_symbol] = (actinium_geometry.ActiniumGeometry(), material)
            elif atom_symbol == "Th":
                self._atoms_parameters[atom_symbol] = (thorium_geometry.ThoriumGeometry(), material)
            elif atom_symbol == "Pa":
                self._atoms_parameters[atom_symbol] = (protactinium_geometry.ProtactiniumGeometry(), material)
            elif atom_symbol == "U":
                self._atoms_parameters[atom_symbol] = (uranium_geometry.UraniumGeometry(), material)
            elif atom_symbol == "Np":
                self._atoms_parameters[atom_symbol] = (neptunium_geometry.NeptuniumGeometry(), material)
            elif atom_symbol == "Pu":
                self._atoms_parameters[atom_symbol] = (plutonium_geometry.PlutoniumGeometry(), material)
            elif atom_symbol == "Am":
                self._atoms_parameters[atom_symbol] = (americium_geometry.AmericiumGeometry(), material)
            elif atom_symbol == "Cm":
                self._atoms_parameters[atom_symbol] = (curium_geometry.CuriumGeometry(), material)
            elif atom_symbol == "Bk":
                self._atoms_parameters[atom_symbol] = (berkelium_geometry.BerkeliumGeometry(), material)
            elif atom_symbol == "Cf":
                self._atoms_parameters[atom_symbol] = (californium_geometry.CaliforniumGeometry(), material)
            elif atom_symbol == "Es":
                self._atoms_parameters[atom_symbol] = (einsteinium_geometry.EinsteiniumGeometry(), material)

        atom_3d = Atom3D(*self._atoms_parameters[atom_symbol])
        return atom_3d








import numpy as np

from PenguinMol3D.geometries.shapes.base_parametric_geometry import BaseParametricGeometry


class Ellipsoid(BaseParametricGeometry):
    def __init__(self,
                 width: int = 1,
                 height: int = 1,
                 depth: int = 1,
                 radius_segments: int = 32,
                 height_segments: int = 32,
                 colors: list[list] = [[1., 0., 0.],
                                       [0., 1., 0.],
                                       [0., 0., 1.],
                                       [0., 1., 1.],
                                       [1., 0., 1.],
                                       [1., 1., 0.]]
                 ):

        def s(u,v):
            return [ width/2 * np.sin(u) * np.cos(v),
                    height/2 * np.sin(v),
                     depth/2 * np.cos(u) * np.cos(v)]

        BaseParametricGeometry.__init__(self,
                                        0,
                                        2 * np.pi,
                                        radius_segments,
                                        -np.pi / 2,
                                        np.pi / 2,
                                        height_segments,
                                        s,
                                        colors)


from PenguinMol3D.geometries.base_geometry import BaseGeometry
from PenguinMol3D.operations.circle_operations import ParametricCircle
from PenguinMol3D.operations.pipe_operations import ParametricPipe
from PenguinMol3D.operations.vector_operations import VectorOperations


class BaseBond(ParametricPipe, BaseGeometry):
    def __init__(self,
                 path_points: list[list[float]],
                 path_colors: list[list[float]],
                 contour_radius: float = 0.35,
                 contour_steps: int = 32):

        contour_points = ParametricCircle.build_circle(contour_radius,
                                                       contour_steps)

        ParametricPipe.__init__(self, path_points, contour_points)
        BaseGeometry.__init__(self)

        for c_idx in range(len(self.contours) - 1):
            contour1 = self.contours[c_idx]
            contour2 = self.contours[c_idx + 1]
            normals1 = self.normals[c_idx]
            normals2 = self.normals[c_idx + 1]
            contour_color1 = path_colors[c_idx]
            contour_color2 = path_colors[c_idx + 1]

            for v_idx in range(contour_steps):
                c0v0 = contour1[v_idx]
                c1v0 = contour2[v_idx]
                vn0 = normals1[v_idx]
                vn1 = normals1[v_idx]

                sv_idx = v_idx + 1
                if sv_idx == contour_steps:
                    c0v1 = contour1[0]
                    c1v1 = contour2[0]
                    vn2 = normals1[0]
                    vn3 = normals2[0]
                else:
                    c0v1 = contour1[sv_idx]
                    c1v1 = contour2[sv_idx]
                    vn2 = normals1[sv_idx]
                    vn3 = normals2[sv_idx]

                self.vertex_position.extend([c0v0, c0v1, c1v1,
                                              c0v0, c1v1, c1v0])
                self.vertex_color.extend([contour_color1,
                                           contour_color1,
                                           contour_color2,
                                           contour_color1,
                                           contour_color2,
                                           contour_color2])
                self.vertex_normal.extend([vn0, vn2, vn3, vn0, vn3, vn1])

                fn0 = VectorOperations.calc_normal(c0v0, c1v1, c0v1)
                fn1 = VectorOperations.calc_normal(c0v0, c1v0, c1v1)
                self.face_normal.extend([fn0, fn0, fn0, fn1, fn1, fn1])

        self.num_vertices = len(self._vertex_position)





import numpy as np

from PenguinMol3D.operations.line_operations import ParametricLine


class ParametricPlane:

    def __init__(self,
                 point: list[float],
                 normal: list[float] = [0., 0., 1.]):
        """class for a 3D plane with normal vector (a,b,c) and a point (x0,y0,z0)
           ax + by + cz + d = 0,  where d = -(ax0 + by0 + cz0)"""

        self.p = np.array(point)
        self.normal = np.array(normal)
        self.normal_length = float(np.sqrt(np.dot(self.normal, self.normal)))
        self.d = -float(np.dot(self.normal, self.p))
        self.distance = -self.d / self.normal_length

    @classmethod
    def from_three_points(cls,
                          p1: list[float],
                          p2: list[float],
                          p3: list[float]):

        AB = np.array(p2) - np.array(p1)
        AC = np.array(p3) - np.array(p1)
        normal = np.cross(AB, AC)
        return cls(p1, normal)


    def get_distance(self, point: list[float]) -> np.ndarray:
        """compute the shortest distance from a given point p2 to the plane defined by normal and p1
           Note: The distance is signed. If the distance is negative, the point is in
           opposite side of the plane.

           D = (a * Px + b * Py + c * Pz + d) / sqrt(a*a + b*b + c*c)
           """

        dot = np.dot(self.normal, point)
        return (dot + self.d) / self.normal_length

    def normalize(self):
        """normalize by dividing each coefficient by the length of normal"""
        length_inv = 1. / self.normal_length
        self.normal *= length_inv
        self.normal_length = 1.
        self.d *= length_inv
        self.distance = -self.d

    def get_intersection_line(self, line: ParametricLine) -> np.ndarray:
        """find the intersect point, substitute a point on the line to the plane equation,
           then solve for alpha a point on a line: (x0 + x*t, y0 + y*t, z0 + z*t)
           plane: a*X + b*Y + c*Z + d = 0

           a*(x0 + x*t) + b*(y0 + y*t) + c*(z0 + z*t) + d = 0
           a*x0 + a*x*t + b*y0 + b*y*t + c*z0 + c*z*t + d = 0
           (a*x + b*x + c*x)*t = -(a*x0 + b*y0 + c*z0 + d)

           t = -(a*x0 + b*y0 + c*z0 + d) / (a*x + b*x + c*x)"""

        dot1 = np.dot(self.normal, line.point)
        dot2 = np.dot(self.normal, line.direction)

        if not dot2.any():
            return np.array([np.nan, np.nan, np.nan])

        t = -(dot1 + self.d) / dot2

        return line.point + (line.direction * t)

    def get_intersection_plane(self, plane) -> ParametricLine:
        """find the intersection line of 2 planes
           P1: N1 dot p + d1 = 0 (a1*X + b1*Y + c1*Z + d1 = 0)
           P2: N2 dot p + d2 = 0 (a2*X + b2*Y + c2*Z + d2 = 0)

           L: p0 + a*V where
           V is the direction vector of intersection line = (a1,b1,c1) x (a2,b2,c2)
           p0 is a point, which is on the L and both P1 and P2 as well

           p0 can be found by solving a linear system of 3 planes
           P1: N1 dot p + d1 = 0     (given)
           P2: N2 dot p + d2 = 0     (given)
           P3: V dot p = 0           (chosen where d3=0)

           Use the formula for intersecting 3 planes to find p0;
           p0 = ((-d1*N2 + d2*N1) x V) / V dot V"""

        direction = np.cross(self.normal, plane.normal)
        if not direction.any():
            return ParametricLine(np.array([np.nan, np.nan, np.nan]),
                                  np.array([np.nan, np.nan, np.nan]))

        dot = np.dot(direction, direction)
        n1 = plane.d * self.normal
        n2 = -self.d * plane.normal
        p = np.cross((n1 + n2), direction) / dot

        return ParametricLine(direction, p)

    def is_intersected_line(self, line: ParametricLine) -> bool:
        """determine if it intersects with the line"""
        dot = np.dot(self.normal, line.direction)

        if not dot.any():
            return False
        return True

    def is_intersected_plane(self, plane) -> bool:
        """determine if it intersects with the other plane"""
        cross = np.cross(self.normal, plane.normal)

        if not cross.any():
            return False
        return True





from PenguinMol3D.objects.light.base_light import BaseLight


class Directional(BaseLight):
    def __init__(self,
                 color: list[float]     = [1.,  1., 1.],
                 direction: list[float] = [0., -1., 0.]):

        BaseLight.__init__(self, light_type=BaseLight.DIRECTIONAL)
        self.color = color
        self.set_direction(direction)



import numpy as np
from rdkit.Chem.rdchem import BondType

from PenguinMol3D.general.globals import ELEMENT_COLORS
from PenguinMol3D.geometries.base_geometry_group import BaseGeometryGroup
from PenguinMol3D.geometries.bonds.double_bond_geometry import DoubleBondGeometry
from PenguinMol3D.geometries.bonds.single_bond_geometry import SingleBondGeometry
from PenguinMol3D.geometries.bonds.triple_bond_geometry import TripleBondGeometry
from PenguinMol3D.materials.bond_phong_material import PhongMaterial
from PenguinMol3D.objects.mesh import Mesh
from PenguinMol3D.operations.vector_operations import VectorOperations


def translate_vector(path: np.ndarray, normal: np.ndarray, value):
    """Translates points of a bond path in a direction (axis)
       orthogonal to path direction and normal of three atom's plane
       NOTE: this transformation is nesessary to generate double and triple bonds
    """
    curr_dir = path[0] - path[1]
    curr_dir /= np.linalg.norm(curr_dir)
    axis = np.cross(curr_dir, normal)
    axis /= np.linalg.norm(axis)

    path += (value * axis)

class Bonds3D(Mesh):
    def __init__(self, geometry, material):
        Mesh.__init__(self, geometry, material)


class BondsGeometry(BaseGeometryGroup):
    def __init__(self):
        BaseGeometryGroup.__init__(self)

    def add_bond(self, bond):
        BaseGeometryGroup.add_geometry(self, bond)

class SingleBond(SingleBondGeometry):
    def __init__(self,
                 path_points: list[list[float]],
                 path_colors: list[list[float]],
                 material,
                 normal: np.ndarray = None):

        SingleBondGeometry.__init__(self,
                                    np.array(path_points),
                                    np.array(path_colors))



class DoubleBond(DoubleBondGeometry):
    def __init__(self,
                 path_points: list[list[float]],
                 path_colors: list[list[float]],
                 material,
                 normal: np.ndarray = None):

        f_bond_path = np.array(path_points)
        translate_vector(f_bond_path, normal, 0.1)

        DoubleBondGeometry.__init__(self, f_bond_path, np.array(path_colors))

        s_bond_path = np.array(path_points)
        translate_vector(s_bond_path, normal, -0.1)

        s_bond_geo = DoubleBondGeometry(s_bond_path,
                                        np.array(path_colors))

        self.merge(s_bond_geo)

class TripleBond(TripleBondGeometry):
    """Triple bond mesh"""
    def __init__(self,
                 path_points: list[list[float]],
                 path_colors: list[list[float]],
                 material,
                 normal: np.ndarray = None):

        f_bond_path = np.array(path_points)
        translate_vector(f_bond_path, normal, -0.2)

        TripleBondGeometry.__init__(self, f_bond_path, np.array(path_colors))


        s_bond_path = np.array(path_points)
        s_bond_geo = TripleBondGeometry(s_bond_path,
                                        np.array(path_colors))

        t_bond_path = np.array(path_points)
        translate_vector(t_bond_path, normal, 0.2)
        t_bond_geo = TripleBondGeometry(t_bond_path,
                                        np.array(path_colors))

        self.merge(s_bond_geo)
        self.merge(t_bond_geo)


class BondsFactory:
    """Generates 3D models of bonds"""
    def __init__(self):
        self._bonds = BondsGeometry()
        self._material = PhongMaterial(properties={"use_vertex_colors" : True})
        self._material.settings["cull_face"] = True
        self._material.settings["front_side"] = False
        self._bond_types = {BondType.SINGLE : SingleBond,
                            BondType.DOUBLE : DoubleBond,
                            BondType.TRIPLE : TripleBond}

    def create_bond_geometry(self,
                             path_points: list[list[float]],
                             bond_type: BondType,
                             beg_atom_symbol: str,
                             end_atom_symbol: str,
                             normal: np.ndarray = None
                             ):
        """Generates and returns a 3D model of bond with specified geometry (type) and material"""
        beg_atom_color = ELEMENT_COLORS[beg_atom_symbol]
        end_atom_color = ELEMENT_COLORS[end_atom_symbol]

        curr_dir = np.array(path_points[0]) - np.array(path_points[1])
        curr_dir /= np.linalg.norm(curr_dir)

        midpoint = np.array(VectorOperations.calc_midpoint(path_points[0],
                                                  path_points[1]))

        midpoint1 = midpoint.copy()
        midpoint1 -= (0.0001 * curr_dir)
        midpoint2 = midpoint.copy()
        midpoint2 += (0.0001 * curr_dir)


        bond3d = self._bond_types[bond_type]([path_points[0], midpoint1, midpoint2, path_points[1]],
                                             [beg_atom_color, beg_atom_color, end_atom_color, end_atom_color],
                                             self._material,
                                             normal=normal)


        self._bonds.add_bond(bond3d)

    def get_bonds(self) -> Bonds3D:
        self._bonds.create_attributes()

        for attribute in self._bonds.attributes.values():
            attribute.upload_data()

        return Bonds3D(self._bonds, self._material)











import numpy as np

class VectorOperations:

    @staticmethod
    def calc_midpoint(vec1: list[float], vec2: list[float]) -> list[float]:
        return [(vec1[0] + vec2[0]) / 2,
                (vec1[1] + vec2[1]) / 2,
                (vec1[2] + vec2[2]) / 2]

    @staticmethod
    def calc_normal(p0:  list[float], p1:  list[float], p2:  list[float]) -> np.ndarray:
        v1 = np.array(p1) - np.array(p0)
        v2 = np.array(p2) - np.array(p0)
        normal = np.cross(v1, v2)
        return normal / np.linalg.norm(normal)

    @staticmethod
    def calc_quaternion(axis: np.ndarray | list[float], angle: float) -> np.ndarray:
        axis = axis / np.linalg.norm(axis)
        q = VectorOperations.normalize_quaternion([axis[0] * np.sin(angle / 2),
                                                   axis[1] * np.sin(angle / 2),
                                                   axis[2] * np.sin(angle / 2),
                                                   np.cos(angle / 2)])
        return q

    @staticmethod
    def calc_unit_vector(vec1: np.ndarray, vec2: np.ndarray) -> np.ndarray:
        diff = vec1 - vec2
        return diff / np.linalg.norm(diff)

    @staticmethod
    def calc_vector_length(vec: np.ndarray | list[float]) -> float:
        return float(np.sqrt(np.dot(vec, vec)))

    @staticmethod
    def convert_3x1_to_4x1(vec: np.ndarray | list[float]) -> np.ndarray:
        return np.array([vec[0], vec[1], vec[2], 1.])

    @staticmethod
    def convert_4x1_to_3x1(vec: np.ndarray | list[float]) -> np.ndarray:
        return np.array([vec[0], vec[1], vec[2]])

    @staticmethod
    def normalize_quaternion(vec: np.ndarray | list[float]) -> np.ndarray:
        return vec / np.dot(vec, vec)

    @staticmethod
    def rotate_align_rodrigues(vec1: np.ndarray, vec2: np.ndarray) -> np.ndarray:
        angle = np.arccos(np.dot(vec1, vec2))
        axis = np.cross(vec1, vec2)
        cos_a = np.cos(angle)
        sin_a = np.sin(angle)
        t = 1 - cos_a

        return vec1 * cos_a + np.cross(axis, vec1) * sin_a + axis * (axis * vec1) * t







from OpenGL.GL import *


class OpenGLUtils:
    @staticmethod
    def initialize_shader(shader_code: str, shader_type: str):
        shader_code = f"#version 330 core\n{shader_code}"
        shader_ref = glCreateShader(shader_type)
        glShaderSource(shader_ref, shader_code)
        glCompileShader(shader_ref)

        compile_success = glGetShaderiv(shader_ref, GL_COMPILE_STATUS)
        if not compile_success:
            error_message = glGetShaderInfoLog(shader_ref).decode("utf-8")
            glDeleteShader(shader_ref)
            raise Exception(f"\n{error_message}")

        return shader_ref

    @staticmethod
    def initialize_program(vert_shader_code: str,
                           frag_shader_code: str
                           ):
        vert_shader_ref = OpenGLUtils.initialize_shader(vert_shader_code,
                                                          GL_VERTEX_SHADER)
        frag_shader_ref = OpenGLUtils.initialize_shader(frag_shader_code,
                                                          GL_FRAGMENT_SHADER)
        program_ref = glCreateProgram()

        glAttachShader(program_ref, vert_shader_ref)
        glAttachShader(program_ref, frag_shader_ref)
        glLinkProgram(program_ref)

        link_success = glGetProgramiv(program_ref, GL_LINK_STATUS)
        if not link_success:
            error_message = glGetProgramInfoLog(program_ref).decode("utf-8")
            glDeleteProgram(program_ref)
            raise Exception(f"\n{error_message}")

        return program_ref

    @staticmethod
    def print_system_info():
        print("Vendor: " + glGetString(GL_VENDOR).decode("utf-8"))
        print("Renderer: " + glGetString(GL_RENDERER).decode("utf-8"))
        print("OpenGl version supported: " + glGetString(GL_VERSION).decode("utf-8"))
        print("GLSL version supported: " + glGetString(GL_SHADING_LANGUAGE_VERSION).decode("utf-8"))




from PenguinMol3D.geometries.shapes.ellipsoid import Ellipsoid


class Sphere(Ellipsoid):
    def __init__(self,
                 radius: int = 1,
                 radius_segments: int = 4,
                 height_segments: int = 2,
                 colors: list[list] = [[1., 0., 0.],
                                       [0., 1., 0.],
                                       [0., 0., 1.],
                                       [0., 1., 1.],
                                       [1., 0., 1.],
                                       [1., 1., 0.]]
                 ):

        Ellipsoid.__init__(self,
                           2 * radius,
                           2 * radius,
                           2 * radius,
                           radius_segments,
                           height_segments,
                           colors)


import numpy as np

from PenguinMol3D.geometries.shapes.base_shape_geometry import BaseShapeGeometry


class BaseParametricGeometry(BaseShapeGeometry):
    def __init__(self,
                 u_start: int|float,
                 u_end: int|float,
                 u_resolution: int|float,
                 v_start: int|float,
                 v_end: int|float,
                 v_resolution: int|float,
                 surface_function: int|float,
                 colors: list[list] = [[1., 0., 0.],
                                       [0., 1., 0.],
                                       [0., 0., 1.],
                                       [0., 1., 1.],
                                       [1., 0., 1.],
                                       [1., 1., 0.]]):

        BaseShapeGeometry.__init__(self)

        delta_u = (u_end - u_start) / u_resolution
        delta_v = (v_end - v_start) / v_resolution

        uv_space = []
        vertex_normals = []
        for u_idx in range(u_resolution + 1):
            v_array, v_normal_array = [], []
            for v_idx in range(v_resolution + 1):
                u = u_start + u_idx * delta_u
                v = v_start + v_idx * delta_v
                v_array.append(surface_function(u,v))

                P0 = surface_function(u, v)
                P1 = surface_function(u + 0.0001, v)
                P2 = surface_function(u, v + 0.0001)
                v_normal_array.append(self._calc_normal(P0, P1, P2))
            uv_space.append(v_array)
            vertex_normals.append(v_normal_array)

        """uvs = []
        for u_idx in range(u_resolution + 1):
            v_array = []
            for v_idx in range(v_resolution + 1):
                u = u_idx/u_resolution
                v = v_idx/v_resolution
                v_array.append([u, v])
            uvs.append(v_array)"""


        vertex_positions   = []
        vertex_colors      = []
        """uv_data = []"""
        vertex_normal_data = []
        face_normal_data   = []

        for x_idx in range(u_resolution):
            for y_idx in range(v_resolution):
                V0 = uv_space[x_idx + 0][y_idx + 0]
                V1 = uv_space[x_idx + 1][y_idx + 0]
                V2 = uv_space[x_idx + 0][y_idx + 1]
                V3 = uv_space[x_idx + 1][y_idx + 1]
                vertex_positions += [V0.copy(), V1.copy(), V3.copy(),
                              V0.copy(), V3.copy(), V2.copy()]

                vertex_colors    += colors

                """uv0 = uvs[x_idx + 0][y_idx + 0]
                uv1 = uvs[x_idx + 1][y_idx + 0]
                uv2 = uvs[x_idx + 0][y_idx + 1]
                uv3 = uvs[x_idx + 1][y_idx + 1]

                uv_data += [uv0, uv1, uv3, uv0, uv3, uv2]"""

                n0 = vertex_normals[x_idx + 0][y_idx + 0]
                n1 = vertex_normals[x_idx + 1][y_idx + 0]
                n2 = vertex_normals[x_idx + 0][y_idx + 1]
                n3 = vertex_normals[x_idx + 1][y_idx + 1]

                vertex_normal_data += [n0,n1,n3,n0,n3,n2]

                fn0 = self._calc_normal(n0,n1,n3)
                fn1 = self._calc_normal(n0,n3,n2)
                face_normal_data += [fn0, fn0, fn0, fn1, fn1, fn1]


        self.add_attribute(vertex_positions, "vec3", "vertex_position")
        self.add_attribute(vertex_colors,    "vec3", "vertex_color")
        """self.add_attribute(vertex_positions, "vec2", "uv_data")"""
        self.add_attribute(vertex_normal_data, "vec3", "vertex_normal")
        self.add_attribute(face_normal_data, "vec3", "face_normal")
        self.count_vertices()

    def _calc_normal(self, P0, P1, P2) -> np.ndarray:
        v1 = np.array(P1) - np.array(P0)
        v2 = np.array(P2) - np.array(P0)
        normal = np.cross(v1, v2)
        normal /= np.linalg.norm(normal)
        return normal


import os
import glfw
import glfw.GLFW as GLFW_CONSTANTS
from OpenGL.GLUT import *
from rdkit.Chem import MolFromMolFile, Kekulize

from PenguinMol3D.general.renderer import Renderer
from PenguinMol3D.objects.camera import Camera
from PenguinMol3D.objects.light.ambient import Ambient
from PenguinMol3D.objects.light.directional import Directional
from PenguinMol3D.objects.light.point import Point
from PenguinMol3D.objects.mol_3d import Mol3D
from PenguinMol3D.objects.molecular_scene import MolecularScene
from PenguinMol3D.objects.trackball import Trackball


class GLFWScreenshotExample:
    def __init__(self):
        self.title = "PenguinMol3D"
        self.width = 1000
        self.height = 800
        self.window = None
        self.running = True
        self.prev_pos = None

        self.make_glfw_window()
        self.make_scene()

    def make_glfw_window(self):
        """Create window and grab OpenGL context"""
        glfw.init()
        glfw.window_hint(GLFW_CONSTANTS.GLFW_CLIENT_API, GLFW_CONSTANTS.GLFW_OPENGL_API)
        glfw.window_hint(GLFW_CONSTANTS.GLFW_RESIZABLE, GLFW_CONSTANTS.GLFW_FALSE)
        glfw.window_hint(GLFW_CONSTANTS.GLFW_CONTEXT_VERSION_MAJOR, 3)
        glfw.window_hint(GLFW_CONSTANTS.GLFW_CONTEXT_VERSION_MINOR, 1)
        glfw.window_hint(GLFW_CONSTANTS.GLFW_SAMPLES, 4)
        glfw.window_hint(GLFW_CONSTANTS.GLFW_DEPTH_BITS, 24)

        self.window = glfw.create_window(self.width, self.height, self.title, None, None)

        glfw.set_key_callback(self.window, self.key_callback)
        glfw.set_cursor_pos_callback(self.window, self.mouse_cursor_callback)
        glfw.set_window_close_callback(self.window, self.close_callback)

        glfw.make_context_current(self.window)

    """GLFW specific methods to handle events"""
    def close_callback(self, window):
        self.running = False

    def mouse_cursor_callback(self, window, xpos: float, ypos: float):
        left_mouse_button = glfw.get_mouse_button(window, GLFW_CONSTANTS.GLFW_MOUSE_BUTTON_LEFT)
        if left_mouse_button == GLFW_CONSTANTS.GLFW_PRESS:
            if self.prev_pos:
                """Use trackball to convert previuous and current cursor coordinates into rotational matrix"""
                rotmat = Trackball.simulate_trackball([self.width, self.height],
                                                      [xpos, ypos],
                                                      [self.prev_pos[0], self.prev_pos[1]])
                self.mol_3d.apply_transform(rotmat, local_coord=False)
            self.prev_pos = (xpos, ypos)

        right_mouse_button = glfw.get_mouse_button(window, GLFW_CONSTANTS.GLFW_MOUSE_BUTTON_RIGHT)
        if right_mouse_button == GLFW_CONSTANTS.GLFW_PRESS:
            if self.prev_pos:
                """Scale object to achieve effect similar to zoom in/zoom out"""
                if self.prev_pos[1] < ypos:
                    self.mol_3d.scale(1.1, local_coord=True)
                else:
                    self.mol_3d.scale(0.9, local_coord=True)
            self.prev_pos = (xpos, ypos)

        if left_mouse_button == GLFW_CONSTANTS.GLFW_RELEASE and \
                right_mouse_button == GLFW_CONSTANTS.GLFW_RELEASE:
            self.prev_pos = None

    def key_callback(self, window, key: int, scancode: int, action: int, mods):
        if key == GLFW_CONSTANTS.GLFW_KEY_ESCAPE and action == GLFW_CONSTANTS.GLFW_PRESS:
            self.running = False

        if key == GLFW_CONSTANTS.GLFW_KEY_S and action == GLFW_CONSTANTS.GLFW_PRESS:
            self.renderer.save_frame(self.width, self.height, "penguinone.png")

    def make_scene(self):
        """Create the Renderer object and setup the Scene"""
        self.renderer = Renderer()
        self.scene = MolecularScene()
        self.camera = Camera(angle_of_view=60,
                             aspect_ratio=self.width / self.height,
                             far=100)

        self.ambient = Ambient(color=[0.1, 0.1, 0.1])
        self.scene.add_child(self.ambient)

        self.directional = Directional(color=[1.0, 1.0, 1.0],
                                       direction=[-1., -1., -2.])

        self.scene.add_child(self.directional)

        self.point = Point(position=[2., 2., 2.])
        self.scene.add_child(self.point)

        """Load molecular data from file using one of the RDKit functions"""
        mol_rdkit = MolFromMolFile(os.path.join(os.path.dirname(os.path.abspath(__file__)), "penguinone.sdf"),
                                   sanitize=True,
                                   removeHs=False)
        Kekulize(mol_rdkit, clearAromaticFlags=True)

        """Pass rdkit Mol object as an argument to Mol3D constructor"""
        self.mol_3d = Mol3D(mol_rdkit)
        self.scene.add_molecule(self.mol_3d)
        self.camera.set_bb_based_position(self.mol_3d.bounding_box)

    def run_application(self):
        """Run application event loop until user decides to quit"""
        while self.running:
            self.renderer.render(self.scene, self.camera)
            glfw.swap_buffers(self.window)
            glfw.poll_events()

        glfw.destroy_window(self.window)
        glfw.terminate()

if __name__ == "__main__":
    app = GLFWScreenshotExample()
    app.run_application()




import os
import setuptools

with open(os.path.join("PenguinMol3D", "README.md"), "r", encoding = "utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name = "PenguinMol3D",
    version = "0.0.4",
    author = "Yevhen Kustovskiy",
    author_email = "ykustovskiy@gmail.com",
    description = "3D visualization of small compounds",
    long_description = long_description,
    long_description_content_type = "text/markdown",
    project_urls = {
        "Homepage": "https://github.com/YevhenKustovskiy/penguinemol3d",
    },
    classifiers = [
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License v2 (GPLv2)",
        "Operating System :: OS Independent",
    ],
    packages = [
    "PenguinMol3D", 
    "PenguinMol3D.factories",
    "PenguinMol3D.general",
    "PenguinMol3D.geometries",
    "PenguinMol3D.geometries.atoms",
    "PenguinMol3D.geometries.bonds",
    "PenguinMol3D.geometries.shapes",
    "PenguinMol3D.materials",
    "PenguinMol3D.objects",
    "PenguinMol3D.objects.light",
    "PenguinMol3D.operations"
    ],
    package_data={
      'PenguinMol3D': ["*.txt", "*.sdf", "*.md"],
    },
    include_package_data=True,
    install_requires=[
    "numpy", 
    "rdkit", 
    "PyOpenGL", 
    "PyOpenGL_accelerate", 
    "glfw",
    "Pillow"
    ],
    python_requires = ">=3.10"
)


from OpenGL.GL import *


class Uniform:
    def __init__(self, data: int|float|list, dtype: str):
        self._data = data
        self._dtype = dtype
        self._variable_ref = None

    def __iadd__(self, other):
        self._data += other
        return self

    def __isub__(self, other):
        self._data -= other
        return self

    @property
    def data(self):
        return self._data

    @data.setter
    def data(self, data):
        self._data = data

    def locate_variable(self, program_ref, variable_name):
        if self._dtype == "Light":
            self._variable_ref = {}
            self._variable_ref["light_type"] = \
                glGetUniformLocation(program_ref, f"{variable_name}.light_type")
            self._variable_ref["color"] = \
                glGetUniformLocation(program_ref, f"{variable_name}.color")
            self._variable_ref["direction"] = \
                glGetUniformLocation(program_ref, f"{variable_name}.direction")
            self._variable_ref["position"] = \
                glGetUniformLocation(program_ref, f"{variable_name}.position")
            self._variable_ref["attenuation"] = \
                glGetUniformLocation(program_ref, f"{variable_name}.attenuation")
        else:
            self._variable_ref = glGetUniformLocation(program_ref, variable_name)

    def upload_data(self):
        if self._variable_ref == -1:
            return

        if self._dtype == "int":

            glUniform1i(self._variable_ref,
                        self._data)
            return

        if self._dtype == "bool":

            glUniform1i(self._variable_ref,
                        self._data)
            return

        if self._dtype == "float":

            glUniform1f(self._variable_ref,
                        self._data)
            return

        if self._dtype == "vec2":

            glUniform2f(self._variable_ref,
                        self._data[0],
                        self._data[1])
            return

        if self._dtype == "vec3":

            glUniform3f(self._variable_ref,
                        self._data[0],
                        self._data[1],
                        self._data[2])
            return

        if self._dtype == "vec4":

            glUniform4f(self._variable_ref,
                        self._data[0],
                        self._data[1],
                        self._data[2],
                        self._data[3])
            return

        if self._dtype == "mat4":

            glUniformMatrix4fv(self._variable_ref,
                               1,
                               GL_TRUE,
                               self._data)
            return

        if self._dtype == "Light":
            glUniform1i(self._variable_ref["light_type"],
                        self._data.light_type)

            glUniform3f(self._variable_ref["color"],
                        self._data.color[0],
                        self._data.color[1],
                        self._data.color[2])

            direction = self._data.get_direction()
            glUniform3f(self._variable_ref["direction"],
                        direction[0],
                        direction[1],
                        direction[2])

            position = self._data.get_position()
            glUniform3f(self._variable_ref["position"],
                        position[0],
                        position[1],
                        position[2])

            glUniform3f(self._variable_ref["attenuation"],
                        self._data.attenuation[0],
                        self._data.attenuation[1],
                        self._data.attenuation[2])
            return

        raise Exception(f"Variable has unknown type {self._dtype}")



from PenguinMol3D.objects.base_object_3d import BaseObject3D


class BaseLight(BaseObject3D):
    AMBIENT = 1
    DIRECTIONAL = 2
    POINT = 3
    def __init__(self, light_type: int = 0):
        BaseObject3D.__init__(self)
        self._light_type  = light_type
        self._color       = [1., 1., 1.]
        self._attenuation = [1., 0., 0.]

    @property
    def light_type(self) -> int:
        return self._light_type

    @light_type.setter
    def light_type(self, light_type: int):
        self._light_type = light_type

    @property
    def color(self) -> list[float]:
        return self._color

    @color.setter
    def color(self, color: list[float]):
        self._color = color

    @property
    def attenuation(self) -> list[float]:
        return self._attenuation

    @attenuation.setter
    def attenuation(self, attenuation: list[float]):
        self._attenuation = attenuation



from OpenGL.GL import *

from PenguinMol3D.materials.base_material import BaseMaterial


class PhongMaterial(BaseMaterial):
    def __init__(self, texture=None, properties={}):
        vert_shader_code = """
                           uniform mat4 projection_matrix;
                           uniform mat4 view_matrix;
                           uniform mat4 model_matrix;
                           uniform vec3 base_color;
                           uniform bool use_vertex_colors;
                           in vec3 vertex_position;
                           in vec3 vertex_normal;
                           in vec3 vertex_color;
                           out vec3 position;
                           out vec3 normal;
                           out vec3 color;
                           
                           void main()
                           {
                               gl_Position = projection_matrix * 
                                             view_matrix * 
                                             model_matrix *
                                             vec4(vertex_position, 1);
                               position = vec3(model_matrix * vec4(vertex_position, 1));
                               normal = normalize(mat3(model_matrix) * vertex_normal);
                               if (use_vertex_colors)
                               {
                                   color = vertex_color;
                               }
                               else
                               {
                                   color = base_color;
                               }
                                   
                           }
                           """

        frag_shader_code = """
                           struct Light
                           {
                               int light_type;
                               vec3 color;
                               vec3 direction;
                               vec3 position;
                               vec3 attenuation;
                           };
                           
                           uniform Light light0;
                           uniform Light light1;
                           uniform Light light2;
                           uniform Light light3;
                           uniform vec3 view_position;
                           uniform float specular_strength;
                           uniform float shininess;
                           
                           vec3 light_calc(Light light, vec3 point_position, vec3 point_normal)
                           {
                               float ambient     = 0.0;
                               float diffuse     = 0.0;
                               float specular    = 0.0;
                               float attenuation = 1.0;
                               vec3 light_direction = vec3(0.0, 0.0, 0.0);
                               
                               if (light.light_type == 1)
                               {
                                   ambient = 1;
                               }
                               else if (light.light_type == 2)
                               {
                                   light_direction = normalize(light.direction);
                               }
                               else if (light.light_type == 3)
                               {
                                   light_direction = normalize(point_position - light.position);
                                   float distance = length(light.position - point_position);
                                   attenuation = 1.0 / (light.attenuation[0] + 
                                                        light.attenuation[1] * distance +
                                                        light.attenuation[2] * distance * distance);
                               }
                               
                               if (light.light_type > 1)
                               {
                                   point_normal = normalize(point_normal);
                                   diffuse = max(dot(point_normal, - light.direction), 0.0);
                                   diffuse *= attenuation;
                                   if (diffuse > 0)
                                   {
                                       vec3 view_direction = normalize(view_position-point_position);
                                       vec3 reflect_direction = reflect( light_direction, point_normal);
                                       specular = max(dot(view_direction, reflect_direction), 0.0);
                                       specular = specular_strength * pow(specular, shininess);
                                   }
                               }
                               return light.color * (ambient + diffuse + specular);
                           }
                           in vec3 color;
                           in vec3 position;
                           in vec3 normal;
                           out vec4 frag_color;
                           void main()
                           {
                               vec4 final_color = vec4(color, 1.0);
                               vec3 total = vec3(0., 0., 0.);
                               total += light_calc(light0, position, normal);
                               total += light_calc(light1, position, normal);
                               total += light_calc(light2, position, normal);
                               total += light_calc(light3, position, normal);
                               final_color *= vec4(total, 1.0);
                               frag_color = final_color;
                           }
                           """

        BaseMaterial.__init__(self, vert_shader_code, frag_shader_code)

        self.add_uniform(None, "mat4", "model_matrix")
        self.add_uniform(False, "bool", "use_vertex_colors")
        self.add_uniform([1., 1., 1.], "vec3", "base_color")
        self.add_uniform(None, "Light", "light0")
        self.add_uniform(None, "Light", "light1")
        self.add_uniform(None, "Light", "light2")
        self.add_uniform(None, "Light", "light3")


        self.add_uniform([0., 0., 0.], "vec3", "view_position")
        self.add_uniform(1.0, "float", "specular_strength")
        self.add_uniform(32.0, "float", "shininess")

        self.locate_uniforms()

        self._settings["cull_face"] = False
        self._settings["front_side"] = True
        self._settings["back_side"] = True
        self._settings["wireframe"]   = False
        self._settings["line_width"]  = 1

        self.set_properties(properties)
        self.update_render_settings()

    def update_render_settings(self):

        if not self._settings["cull_face"]:
            glDisable(GL_CULL_FACE)
        else:
            glEnable(GL_CULL_FACE)

        if not self._settings["front_side"]:
            glCullFace(GL_FRONT)

        if not self._settings["back_side"]:
            glCullFace(GL_BACK)

        if self._settings["wireframe"]:
            glPolygonMode(GL_FRONT_AND_BACK, GL_LINE)
        else:
            glPolygonMode(GL_FRONT_AND_BACK, GL_FILL)

        glLineWidth(self._settings["line_width"])



from PenguinMol3D.objects.light.base_light import BaseLight


class Ambient(BaseLight):
    def __init__(self, color: list[float] = [1., 1., 1.]):
        BaseLight.__init__(self,light_type=BaseLight.AMBIENT)
        self.color = color
